#include <stdio.h>
#include <string.h>
#include <stdlib.h>



char *MyFgets(char buf[],int maxn,FILE *fp)
{
	int i;
	char *r=fgets(buf,maxn,fp);
	for(i=0; buf[i]!=0; ++i)
	{
		if(0<=buf[i] && buf[i]<32)
		{
			buf[i]=0;
			break;
		}
	}
	return r;
}

struct DictEntry
{
	struct DictEntry *next;
	unsigned char *fromStr;
	char *toStr;
};

struct Binary
{
	long int sz;
	unsigned char *buf;
};

struct DictEntry *CreateNewDictEntry(const char from[],const char to[])
{
	int fromLen;
	int toLen;
	struct DictEntry *newEntry;

	fromLen=strlen(from);
	toLen=strlen(to);
	newEntry=(struct DictEntry *)malloc(sizeof(struct DictEntry));

	newEntry->next=NULL;
	newEntry->fromStr=(unsigned char *)malloc(fromLen+1);
	newEntry->toStr=(char *)malloc(toLen+1);

	strcpy((char *)(newEntry->fromStr),from);
	strcpy(newEntry->toStr,to);

	if(fromLen<toLen)
	{
		newEntry->toStr[fromLen]=0;
		printf("[%s] is chopped off to [%s].\n",to,newEntry->toStr);
	}


	return newEntry;
}

struct DictEntry *dict[65536];

unsigned int GetRow(struct DictEntry *dictEnt)
{
	unsigned char *ucPtr=(unsigned char *)dictEnt->fromStr;
	unsigned int lower=ucPtr[0];
	unsigned int upper=ucPtr[1];
	return upper*256+lower;
}

void InitDict(void)
{
	int i;
	for(i=0; i<65536; ++i)
	{
		dict[i]=NULL;
	}
}

void AddEntryFromString(const char str[])
{
	int i,row;
	struct DictEntry *newEnt;
	char *from,*to;
	int len=strlen(str);

	from=(char *)malloc(len+1);
	to=(char *)malloc(len+1);
	to[0]=0;

	strcpy(from,str);
	for(i=0; 0!=from[i]; ++i)
	{
		if('$'==from[i] && '$'==from[i+1])
		{
			from[i]=0;
			strcpy(to,str+i+2);
			break;
		}
	}
	if(i<4)
	{
		printf("Matching less than four bytes is too risky.\n");
		printf("May destroy binary.  Skipping.\n");
	}
	else if(0==to[0])
	{
		printf("Skipping empty English text.\n");
	}
	else
	{
		newEnt=CreateNewDictEntry(from,to);
		row=GetRow(newEnt);
		newEnt->next=dict[row];
		dict[row]=newEnt;
	}

	free(from);
	free(to);
}

int ReadDictFile(const char fn[])
{
	FILE *fp=fopen(fn,"r");
	InitDict();
	if(NULL!=fp)
	{
		char buf[256];
		while(NULL!=MyFgets(buf,255,fp))
		{
			if('#'==buf[0])
			{
				continue;
			}
			AddEntryFromString(buf);
		}
		fclose(fp);\
		return 1;
	}
	return 0;
}

struct DictEntry *FindMatch(struct Binary bin,int ptr)
{
	struct DictEntry *found=NULL;
	if(ptr+1<bin.sz)
	{
		unsigned int lower=bin.buf[ptr];
		unsigned int upper=bin.buf[ptr+1];
		unsigned int row=upper*256+lower;
		struct DictEntry *seeker=NULL;
		int curLen=0;

		for(seeker=dict[row]; NULL!=seeker; seeker=seeker->next)
		{
			int i;
			for(i=0; 0!=seeker->fromStr[i] && i+ptr<bin.sz; ++i)
			{
				if(seeker->fromStr[i]!=bin.buf[i+ptr])
				{
					break;
				}
			}

			if(0==seeker->fromStr[i])
			{
				if(NULL==seeker || curLen<i)
				{
					found=seeker;
					curLen=i;
				}
			}
		}
	}
	return found;
}

struct Binary ReadBinary(const char fn[])
{
	struct Binary bin;
	bin.sz=0;
	bin.buf=NULL;
	FILE *fp=fopen(fn,"rb");
	if(NULL!=fp)
	{
		unsigned char *buf;
		long int sz=0;

		fseek(fp,0,SEEK_END);
		sz=ftell(fp);
		fseek(fp,0,SEEK_SET);

		printf("%d bytes.\n",sz);

		buf=(unsigned char *)malloc(sz);
		fread(buf,1,sz,fp);

		fclose(fp);

		bin.sz=sz;
		bin.buf=buf;
	}
	return bin;
}

void WriteBinary(const char fn[],struct Binary bin)
{
	FILE *fp=fopen(fn,"wb");
	if(NULL!=fp)
	{
		int wroteSize=fwrite(bin.buf,1,bin.sz,fp);
		fclose(fp);

		if(wroteSize!=bin.sz)
		{
			printf("File Write Error!\n");
		}
		else
		{
			printf("Wrote %s\n",fn);
		}
	}
	else
	{
		printf("Cannot write to %s\n",fn);
	}
}

void FreeBinary(struct Binary bin)
{
	free(bin.buf);
}

void Englishnize(const char expFn[])
{
	int ptr;
	int ctr=0;
	struct Binary bin=ReadBinary(expFn);
	if(NULL==bin.buf)
	{
		printf("Cannot read %s\n",expFn);
		return;
	}

	printf("Englishnizing %s\n",expFn);

	for(ptr=0; ptr<bin.sz; ++ptr)
	{
		struct DictEntry *match;
		match=FindMatch(bin,ptr);
		if(NULL!=match)
		{
			if(0!=match->toStr[0])
			{
				int i;
				printf("Ptr %d replaced to [%s]\n",ptr,match->toStr);
				for(i=0; 0!=match->fromStr[i]; ++i)
				{
					bin.buf[ptr+i]=' ';
				}
				for(i=0; 0!=match->toStr[i]; ++i)
				{
					bin.buf[ptr+i]=match->toStr[i];
				}
				++ctr;
			}
			else
			{
				printf("Ptr %d matches, but to-str is empty.\n",ptr);
			}
		}
	}

	if(0<ctr)
	{
		printf("%d replaced.\n",ctr);
		printf("Updating file %s\n",expFn);
		WriteBinary(expFn,bin);
	}
	else
	{
		printf("No change to the file.\n");
	}

	FreeBinary(bin);
}

int main(int ac,char *av[])
{
	int i;
	char *bin;

	if(0==ReadDictFile("textlist.txt"))
	{
		printf("Cannot read the dictionary file.\n");
		return 1;
	}

	if(ac<2)
	{
		printf("Usage: run386 -nocrt english.exp <EXP-File>\n");
		return 1;
	}

	for(i=1; i<ac; ++i)
	{
		Englishnize(av[i]);
	}

	return 0;
}
